open Signatures

(* An implementation of identifiers as strings.

   The subset of ``base'' identifiers is the set of all strings that
   do not end with ("__" [ '0'-'9' ]+). So, [basename] chops off the
   longest (repeated) suffix of this form, while [combine] appends
   "__", followed with a representation of the integer salt grain, to
   the basename. *)

module String = struct

  type t = string

  let compare = String.compare

  let mirror s =
    let n = String.length s in
    let t = String.create n in
    for i = 0 to n-1 do
      t.[i] <- s.[n-1-i]
    done;
    t

  let basename s =
    mirror (Chop.chop (Lexing.from_string (mirror s)))

  let combine s i =
    if i = 0 then
      s
    else
      s ^ "__" ^ (string_of_int i)

  module Map = Map.Make(String)

end

(* An implementation of atoms. *)

module Make (Identifier : Identifier) = struct

  type identifier = Identifier.t

  module Atom = struct

    (* An atom is implemented as a pair of an integer, the atom's
       identity, and an identifier, the atom's basename.

       We make identifiers part of atoms, instead of storing them in a
       separate table, so as to avoid the need for collecting garbage in
       such a table. *)

    type t = {
	identity: int;
	basename: Identifier.t
      }

    (* An integer counter holds the next available atom. *)

    let counter =
      ref 0

    let allocate () =
      let number = !counter in
      counter := number + 1;
      assert (number >= 0);
      number

    (* [freshb identifier] returns a fresh atom with basename [basename
       identifier]. *)

    let freshb identifier = {
      identity = allocate();
      basename = Identifier.basename identifier
    }

    (* [fresha a] returns a fresh atom whose basename is that of the
       atom [a]. *)

    let fresha a = {
      identity = allocate();
      basename = a.basename
    }

    (* Comparison over atoms. *)

    let equal a b =
      a.identity = b.identity

    let compare a b =
      a.identity - b.identity

    let hash a =
      Hashtbl.hash a.identity

    (* [identity a] maps every atom to a distinct integer. This number
       can be viewed as the atom's identity and can be turned into a
       unique name for printing purposes, if desired. *)

    let identity a =
      a.identity

    (* An atom can be queried about its base identifier. *)

    let basename a =
      a.basename

    exception Unknown of t

  end

  open Atom

  (* Efficient implementations of maps and sets over integers. *)

  module IntMap = Patricia.Big
  module IntSet = IntMap.Domain

  module AtomSet = struct

    (* Sets of atoms are represented as maps of integers to atoms. *)

    type element =
	Atom.t

    type t =
	element IntMap.t

    let empty =
      IntMap.empty

    let add a s =
      IntMap.add a.identity a s

    let singleton a =
      add a empty

    let union =
      IntMap.union

    let mem a s =
      try
        let _ = IntMap.lookup a.identity s in
	true
      with Not_found ->
	false

    let inter s1 s2 =
      IntMap.fold (fun identity a s1 ->
	try
	  let _ = IntMap.lookup identity s2 in
	  IntMap.add identity a s1
	with Not_found ->
	  s1
      ) s1 empty

    let is_empty =
      IntMap.is_empty

    let disjoint s1 s2 =
      IntMap.Domain.disjoint (IntMap.domain s1) (IntMap.domain s2)

    let compare s1 s2 =
      IntMap.compare (fun a b ->
	assert (equal a b);
	0
      ) s1 s2

    let equal s1 s2 =
      compare s1 s2 = 0

    let subset s1 s2 =
      IntMap.Domain.subset (IntMap.domain s1) (IntMap.domain s2)

    let fold f s accu =
      IntMap.fold (fun _ a accu ->
	f a accu
      ) s accu

    let iter f s =
      IntMap.iter (fun _ a ->
	f a
      ) s

    let diff s1 s2 =
      IntMap.corestrict s1 (IntMap.domain s2)

    let cardinal =
      IntMap.cardinal

    let choose s =
      let _, a = IntMap.choose s in
      a

  end

  module AtomMap = struct

    (* Maps over atoms are represented as maps over integers. The
       image of an integer [i] through the map is a pair of the
       atom whose identity is [i] and of the actual data. *)

    (* TEMPORARY implmenter lift pour transformer un ensemble
       en un map; exige dans GMap un map qui donne accs aux clefs? *)
    (* TEMPORARY implmenter restrict/corestrict? *)

    type key =
	Atom.t

    type 'a t =
	(key * 'a) IntMap.t

    let empty =
      IntMap.empty

    let is_empty =
      IntMap.is_empty

    let singleton a v =
      IntMap.singleton a.identity (a, v)

    let add a v m =
      IntMap.add a.identity (a, v) m

    exception Strict of key

    let strict_add a v m =
      try
	IntMap.strict_add a.identity (a, v) m
      with IntMap.Unchanged ->
	raise (Strict a)

    let union =
      IntMap.union

    let lookup a m =
      let _, v = IntMap.lookup a.identity m in
      v

    let find =
      lookup

    let map f m =
      IntMap.map (fun (a, v) -> (a, f v)) m

    let iter f m =
      IntMap.iter (fun _ (a, v) ->
	f a v
      ) m

    let fold f m accu =
      IntMap.fold (fun _ (a, v) accu ->
	f a v accu
      ) m accu

    let mapi f m =
      fold (fun a v m ->
	add a (f a v) m
      ) m empty

    let cardinal =
      IntMap.cardinal

    let choose m =
      let _, binding = IntMap.choose m in
      binding

  end

  module AtomIdMap = struct

    type key = Atom.t

    (* A map of atoms to identifiers is implemented as a map of atoms
       to identifiers, together with a map of basenames to so-called
       salt grains (integer counters). *)

    type t = {
	mapping: Identifier.t AtomMap.t;
	salt: int Identifier.Map.t
      }

    (* The empty map. *)

    let empty = {
      mapping = AtomMap.empty;
      salt = Identifier.Map.empty
    }

    (* The map [add a m] extends [m] by mapping [a] to a unique
       identifier. [a]'s basename is used in the computation of
       this identifier. *)

    let add a m =
      let basename = a.basename in
      let salt = m.salt in
      let grain =
	try
	  Identifier.Map.find basename salt
	with Not_found ->
	  0
      in
      try
	{
	  mapping = AtomMap.strict_add a (Identifier.combine basename grain) m.mapping;
	  salt = Identifier.Map.add basename (grain + 1) salt
        }
      with AtomMap.Strict _ ->
	assert false (* [a] is already a member of the domain of [m]. *)

    (* [add_set s m] extends [m] with a mapping of [a] to a unique
       identifier for every atom [a] in the set [s]. *)

    let add_set s m =
      AtomSet.fold add s m

    (* [lookup a m] is the identifier associated with [a] in the
       map [m]. [Atom.Unknown] is raised if [a] is not a member of
       the domain of [m]. *)

    let lookup a m =
      try
	AtomMap.lookup a m.mapping
      with Not_found ->
	raise (Atom.Unknown a)

    let find =
      lookup

    let iter f m =
      AtomMap.iter f m.mapping

    let fold f m accu =
      AtomMap.fold f m.mapping accu

  end

  module Subst = struct

    type t = Atom.t AtomMap.t

    let id = AtomMap.empty
    let is_id = AtomMap.is_empty
    let singleton = AtomMap.singleton
    let add = AtomMap.add
    let union = AtomMap.union

    let lookup a subst =
      try
	AtomMap.lookup a subst
      with Not_found ->
	a

    let find =
      lookup

    let compose subst1 subst2 =

      (* If a mapping of [a] to [b] appears in [subst2], then a
	 mapping of [a] to [lookup b subst1] must appear in the
	 composition. This is the meaning of the second operand of
	 [union]. For atoms [a] that are not in the domain of
	 [subst2], we keep their image through [subst1]. This is the
	 meaning of the first operand. *)

      IntMap.union subst1 (IntMap.endo_map (fun (a, b) -> (a, lookup b subst1)) subst2)

    let freshen s subst =
      AtomSet.fold (fun a subst ->
	add a (Atom.fresha a) subst
      ) s subst

  end

end

